// 			Fragile Base Class problem and solutions
//
// An implementation of the thread-safe Strstream.
// This new Strstream functionality is distributed by Vendor A
// as a compiled "patch", strstream_v2.o; or it may be rolled in
// into a new version of his "C++ stream library"

#include "strstream_v2.h"

				// Exclusive locks, POSIX style
				// The interface literally follows POSIX (and OSF DCE)
void pthread_mutex_init(int *,int);
void pthread_mutex_destroy(int*);
int pthread_mutex_lock(int*);
int pthread_mutex_unlock(int*);


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

strstream_v2::strstream_v2(void)
{
  pthread_mutex_init(&mutex,0);
}

strstream_v2::~strstream_v2(void)
{
  pthread_mutex_destroy(&mutex);
}

					// freeze (when n != 1) or thaw (when n==0) the stream,
					// asserting or releasing the exclusive lock
void strstream_v2::freeze(const int n)
{
  if(n)
    lock(true), Strstream::freeze(1);
   else Strstream::freeze(0), lock(false);
}

void  strstream_v2::lock(const bool onoff)
{
  assert( (onoff ? pthread_mutex_lock : pthread_mutex_unlock)(&mutex) == 0);
}

//---------------------------------------------------------------------------------
//			"Implementation" of the POSIX-style locks
// The only thing we care in the present project is whether these functions are
// actually called. Still, we follow the POSIX interface, and do some sanity
// checking

void pthread_mutex_init(int * mutex_p,int attributes)
{
  fprintf(stderr,"creating a mutex id=%d\n",(*mutex_p=rand()));
}

void pthread_mutex_destroy(int * mutex_p)
{
  assert(*mutex_p > 0);			// make sure the mutex is not locked when destroyed
  fprintf(stderr,"destroying mutex id=%d\n",*mutex_p);
}

int pthread_mutex_lock(int * mutex_p)
{
  assert(*mutex_p > 0);			// make sure the mutex is not locked already
  fprintf(stderr,"locking mutex id=%d\n",*mutex_p);
  *mutex_p = - (*mutex_p);		// Reverse the sign to indicate the mutex is locked
  return 0;						// return with no error
}

int pthread_mutex_unlock(int * mutex_p)
{
  assert(*mutex_p < 0);			// make sure the mutex has been locked already
  *mutex_p = - (*mutex_p);		// Reverse the sign to indicate the mutex is released
  fprintf(stderr,"unlocking mutex id=%d\n",*mutex_p);
  return 0;						// return with no error
}
